% Songyot Nakariyakul
% March 7, 2006
% Adaptive Branch and Bound Algorithm


function [kFtrIndex, numJ, boundB] = AdaptiveBranchBound(Input, groundtruth, m);


% Adaptive Branch and Bound (ABB) algorithm to select the optimal m features out of n 
% features with the min solution tree but no right-left search strategy b/c it is slow.
% ABB includes 4 new properties 
% I. Node Ordering by SigOrder function 
% II. Obtaining a large initial bound by the SFFS function 
% III. Kmin is the initial starting search level
% IV. Use of predicted Jump levels in searching the tree

% Reference for the ABB alg: Pattern Recognition Letters 28(12), pp. 1415-1427, Sep 2007 

% Input is an Sxn input matrix, where S is the number of samples and n is the total number
%     of features. Each row of the input is thus the response at a given pixel sample.
%     These are the training set data.
% groundtruth is an Sx1 ground truth vector,
%     1 is for class 1 and 2 is for class 2.
% m is the number of features to be selected.

% kFtrIndex is the list of the optimal m features.
% numJ is the number of J calculations.
% boundB is the max J value for the optimal m features. 

% Refer to pseudocode in the above reference.

% LIST(a,b,A) is the 3-D matrix that holds the feature nodes in the tree being considered 
%   at the present time in the tree search. For a given A, an axb matrix array of data 
%   exists, a given column of this array (a specific "b" choice) denotes a path to the 
%   nodes being considered at the present level of the tree. The value of 'a' is the 
%   number of levels of the tree we just jumped by in the search and 'b' denotes the 
%   number of nodes being considered at the present tree level (e.g. the number of 
%   successor nodes of the prior node at a level further up the tree with a large J 
%   value). We now discuss A. The value of A is a 'jump index' that denotes the number of
%   jumps needed to reach the current level. Backtracking jumps (by 1) are subtracted 
%   off; i.e. When backtracking occurs, we decrement A by 1. This LIST stores the paths 
%   to nodes just jumped to. These nodes/paths have to be searched or cut-off. When all 
%   nodes in the tree have been checked or cut off, A = 0 and the search terminates. When 
%   a given node has a J < B and is cut off, that column of LIST is removed.  At the end 
%   of the search, the set of features to be removed to obtain the optimal m features is
%   known, and kFtrIndex is the list of the optimal m features.





% Level is the current level in the tree being evaluated, A < Level.
% IND(A,:) lists # of nodes that must be checked at the present level determined by the 
%    jump-index A (# of nodes is just value of b)  
% dest is the predicted tree level for each jump
% J(A,i) is the J value at node i at jump-index A 
% List(:,:) generates the lists of successor nodes of the current node at the next jump level. 


% step 1. initialization  (Steps refer to steps in the pseudocode in the ref)
n = size(Input, 2);    % n is the # of features
[ftrIndex, FTR, numJ, maxJ] = SFFS(Input, groundtruth, m); % use SFFS to compute the initial bound
ftrIndex(1:m) =  FTR(m,1:m);  % ftrIndex is the current optimal feature set corresponding to the initial bound
boundB = maxJ(m);  % boundB is the initial bound
[ftrOrder, intD, Jall] = SigOrder(Input, groundtruth, boundB);  %use SigOrder to order the feature set

Sample = Input(:,ftrOrder);


%Zmax is the set of n-m features that needs to be removed from the original feat set to get the max J value
mem = ismember(ftrOrder,ftrIndex(1:m));  
Zmax = find(mem == 0); % set initial Zmax to be the original feature set = m SFFS features 

numJ = 0;   % numJ is the number of J calculations
Level = intD %Level is the tree level; set Level = intD (the starting search level)

%Level = 40  %set Level = 40 if there is a memory problem  (uncomment this line)

% step 2. generate successor nodes with the predicted JUMP level
C = nchoosek(1:intD+m,m)'; 
[a, b] = size(C)
a1 = find(C(a,:)~=m+intD);

for j = a1
    Jtmp = computeJ(Sample(:,C(:,j)'), groundtruth); numJ = numJ+1;
    if Jtmp > boundB        % update boundB if Jtmp is larger
        Zmax = setdiff([1:n],C(:,j));
        boundB = Jtmp;
    end
end
C(:,a1) = [];
[a, b] = size(C);
C = [C; repmat([intD+m+1:n]',1,b)];

A = 1;

for j = 1:b
    J(A,j) = computeJ(Sample(:,C(:,j)'), groundtruth); numJ = numJ+1;
    List(j,1:intD) = setdiff([1:n],C(:,j)');
end


clear C


[J, tmp] = sort(J);
List = List';
LIST1 = List(:,tmp);
LIST1 = fliplr(LIST1);
J1(A,:) = fliplr(J(A,:));
    
clear J List  % clear the memory

for jind = 1:b  
    if J1(1,jind) < boundB, break; end
    Z(1:intD) = LIST1(1:intD,jind);
        
    A = 1;   % A is the index (number) of jump levels being evaluated    
    Level = intD;
           
    % This is our jump search prediction mechanism
    beta = log(1-J1(1,jind)/Jall)/log(Level/n);
    if beta == 0; beta = 1; end
    dest = ceil(n*(1-boundB/Jall)^(1/beta));
    if dest > n-m; dest = n-m; end
          
    %dest = Level +1;   %uncomment this line if want no jump search
           
    List = nchoosek(Z(1,Level)+1:dest+m,dest-Level)';
    Level = dest;

    [a, b] = size(List);
    LIST(1:a,1:b,A) = List;    % update LIST with new branch pathes at jump A
    IND(A,1:b) = 1:b;     % update IND at jump A
           
           
    Z1 = Z(1:Level-a);
    for j = 1:b

        tmp = List(:,j);
        Ztmp = [Z1 tmp'];
                            
        if List(end,j) == m+Level  % else check for a single path (B.Yuan)
            
            Ztmp = [Ztmp m+Level+1:n];
            ftrs= setdiff([1:n],Ztmp);
            Jtmp = computeJ(Sample(:,ftrs), groundtruth);  numJ = numJ+1;
            J(A,j) = 0;  IND(A,j) = 0;
            if Jtmp > boundB, Zmax = Ztmp; boundB = Jtmp; end

        else  %else compute J for each successor nodes for jump
                 
            ftrs = setdiff([1:n],Ztmp);
            J(A,j) = computeJ(Sample(:,ftrs), groundtruth); numJ = numJ+1;
        end
    end
        
    % step 3. select new unchecked nodes or nodes not cut off to be searched next
    while A ~= 0  %if JUMP level = 0, terminate the algorithm
        if sum(IND(A,:)) == 0 %if List(i) is empty (all elements of IND(A,:) = 0), all successor nodes are explored, go to step 5.
            J(A,:) = 0;   % set J values of all explored successor nodes to be 0
            Level = Level - nnz(LIST(:,1,A)); % update Level to be the previous Jump level
            LIST(:,:,A) = 0;  % then set LIST(:,:,A) = 0 (all successor nodes are explored)
            A = A-1;  % decrease the number of jump by 1, go to step 5;
        else
            % find the node at a given level that, when removed, gives the max J
            J0 = max(J(A,:));
            z = find(J(A,:)==J0);
       
            a = nnz(LIST(:,1,A));
            %Z is a vector that keeps track of the active branch that is being evaluated
            Z(Level-a+1:Level)= LIST(1:a,z(1),A)';
       
            IND(A,z(1)) = 0;   %delete the IND element with max J from IND for JUMP A
            J(A,z(1)) = 0; % also set J(A,z(1)) = 0, since it is already evaluated
           
            % step 4. check bound (is J at the present node above B?)
            if J0 <= boundB   % if J0 is less than current bound, subtree under the node is rejected
       
                IND(A,:) = 0;    %%%%% empty the List(i)
                J(A,:) = 0;      %%% empty J
                Level = Level - a;
                LIST(:,:,A) = 0;   %% empty LIST
                A = A-1;  % step 5. This corresponds to backtracking.
           
            elseif Level == n-m % if at the highest level (leaf), go to step 6.
                boundB = J0; Zmax = Z(1:n-m);  % step 6;  update boundB and Zmax
                IND(A,:) = 0;   %%%%% empty the List(i)
                J(A,:) = 0;
                Level = Level - a;
                LIST(:,:,A) = 0;
                A = A-1;   % step 5.
           
            else
           
                A = A+1;   % otherwise, increase the number of JUMP and go to step 2.
           
                % This is our prediction mechanism
                beta = log(1-J0/Jall)/log(Level/n);
                if beta == 0; beta = 1; end
                dest = ceil(n*(1-boundB/Jall)^(1/beta));
                if dest > n-m; dest = n-m; end
          
                %dest = Level +1;   %uncomment this line if we want no jump search
           
                List = nchoosek(Z(1,Level)+1:dest+m,dest-Level)';
                Level = dest;
           
                %List
                [a, b] = size(List);
                LIST(1:a,1:b,A) = List;    % update LIST with new branch pathes at jump A
                IND(A,1:b) = 1:b;     % update IND at jump A
           
                Z1 = Z(1:Level-a);
                for j = 1:b
               
                    tmp = List(:,j);
                    Ztmp = [Z1 tmp'];
        
                    if List(end,j) == m+Level  % else check for a single path (B.Yuan)
                       
                        Ztmp = [Ztmp m+Level+1:n];
                        ftrs= setdiff([1:n],Ztmp);
                        Jtmp = computeJ(Sample(:,ftrs), groundtruth);  numJ = numJ+1;
                        J(A,j) = 0;  IND(A,j) = 0;
                        if Jtmp > boundB, Zmax = Ztmp; boundB = Jtmp; end
                   
                    else  %else compute J for each successor nodes for jump
                                           
                        ftrs= setdiff([1:n],Ztmp);
                        J(A,j) = computeJ(Sample(:,ftrs), groundtruth); numJ = numJ+1;
                    end
                end
           
            end
        end
    end
end

% Zmax contains the features that when removed from the feature set gives the optimal J

ftrs= setdiff([1:n],Zmax);
kFtrIndex = ftrOrder(ftrs);   % kFtrIndex is the list of the optimal m features


