% Songyot Nakariyakul 
% Last modified: 13 August 2005
% Sequential Forward Floating Selection (SFFS) feature selection algorithm

function [ftrIndex, FTR, numJ, maxJ] = SFFS(Input, groundtruth, m);

% reference: S. Nakariyakul and D. Casasent, Contaminant detection on poultry carcasses using
%   hyperspectral data: Part I. Algorithms for selection of individual wavebands, Proc. SPIE, 6761, 2007.
%  

% ftrIndex - The current feature set (m + 7).  We overshoot the alg to m+7 to allow backtracking (see ref).
% FTR is a (m+7)x(m+7) feature set matrix.  
%     Row 1 of FTR represents the best single feature (ignore 0s)
%     Row 2 shows the feature set of two features (ignore 0s), and so on 
% We are interested in FTR(m,1:m) = the feature set of m selected features

% numJ is the total number of J calculations
% maxJ is a 1xn vector, where the first element is the J value of the first feature set
%             The second element is the J value of the feature subset of two features, etc


% Input is an numSxn input matrix, where numS is the number of training set samples and 
%     n is the total number of features.  These are the training set data.
% groundtruth is an Sx1 ground truth vector
%     1 is for class 1 and 2 is for class 2
% m - number of features to be selected

numJ = 0;  % number of J calculations
[n] = size(Input,2);   % n = the total number of features

ftrIndex = 0;  % intialize ftrIndex; it can be any dummy number

% SEQUENTIAL FORWARD FLOATING SELECTION
maxJ = zeros(1,n);   % store max J value as a function of # features
restFtrs = 1:n;   % remaining features that are not selected by SFFS


%this WHILE loop finds the m SFFS features
i = 1;
while i < m+8   %overshoot to m+8 to allow backtracking
   
    tmpJ = [];  %initialize tmpJ to be an empty vector
   
   % add one feature to the selected feature set (SFS step)
   for j = restFtrs  
      tmpFtr = [ftrIndex(1:i-1) j]; 
      J = computeJ(Input(:,tmpFtr), groundtruth);   %compute J for all features
      numJ = numJ+1;   % increment the number of J calculations
      tmpJ = [tmpJ J]; % put all Js in tmpJ 
   end
  
   maxFtr = find(tmpJ==max(tmpJ)); % find the location of feature with the max J
   ftrIndex(i) = restFtrs(maxFtr(1));   %put the feature with max J in ftrIndex
   restFtrs(maxFtr(1)) = []; %empty that feature from restFtrs after being selected
   
   if max(tmpJ) > maxJ(i)   %check if the new maxJ is larger than the current maxJ
       maxJ(i) = max(tmpJ);  % maxJ(i) stores the max J value for the i-th feature set
       FTR(i,1:i) = ftrIndex(1:i); %FTR is a matrix whose row stores the selected SFFS features
   end
   
   
  % check backtracking when i > 2
   while i > 2
       
       tmpJ = [];
       % conditional backtracking: remove the least significant feature from the selected set
       for k = 1:i
           tmpFtr = ftrIndex(1:i);
           tmpFtr(k) = [];
           J = computeJ(Input(:,tmpFtr), groundtruth); numJ = numJ+1;
           tmpJ = [tmpJ J];
       end    
      
       % if removing the least sig feat does not improve maxJ, break the while loop and continue adding feat by SFS
       if max(tmpJ) <= maxJ(i-1), break, end
       
       % if removing the least significant feature from the set improves the maxJ(i-1), do it
       if (max(tmpJ) > maxJ(i-1))
       
           leastFtr = find(tmpJ==max(tmpJ));
           restFtrs(length(restFtrs)+1) = ftrIndex(leastFtr); %put the least sig feature back in the restFtrs
           ftrIndex(leastFtr) = []; %remove the least sig feat from the selected set
                    
           i = i-1;  %decrease the feature index by 1   
           maxJ(i) = max(tmpJ);  %update the new maxJ 
           FTR(i,1:i) = ftrIndex(1:i); %update the new feature set in FTR
       
       end
       
   end
   
   i = i+1; %increase the index by one
end




